#!/usr/bin/env python
# coding: utf-8

# In[1]:


#Import data and libraries
import pandas as pd
import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
from datetime import date, timedelta
import matplotlib.ticker as mtick
import matplotlib.ticker as ticker
import scipy.stats as stats
from sklearn.preprocessing   import StandardScaler
import plotly.express as px
import warnings
warnings.filterwarnings('ignore')
import missingno as msno
from sklearn.cluster import KMeans

from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_selection import SelectFromModel
import numpy as np
from sklearn.datasets import load_boston
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split
from sklearn.feature_selection import RFECV
import matplotlib.pyplot as plt
# pip install hvplot install if don't have
from sklearn import metrics
from sklearn.model_selection import cross_val_score
import hvplot.pandas
#Encoder the categorical variables for nomial features
from sklearn.preprocessing import OneHotEncoder
import category_encoders as ce


# In[2]:


#Import dataset
alldata = pd.read_csv('Guideline_Pricing_Full_Data.csv')
print("Dataset shape:\n",alldata.shape,"\n") 


# # Initial EDA

# In[3]:


alldata.info() #get overall view of the dataset


# In[4]:


alldata['Table Name'].value_counts() #look at two products records


# In[5]:


alldata['Currency'].value_counts() #Get a count for currency info


# In[6]:


filtered_df = alldata[alldata['Currency'].notnull()].reset_index(drop=True) #filter the dataset with currency info


# In[7]:


filtered_df['Country'].value_counts()  #Get a count for country info


# In[8]:


#filter the dataset with UNITED STATES in country as this project main focus
filtered_df = filtered_df.loc[filtered_df['Country'] == 'UNITED STATES'].reset_index(drop=True)


# In[9]:


filtered_df['Stage (group)'].value_counts() #look at the closed opportunities


# In[10]:


filtered_df['Stage (group)'].unique() #look at the closed opportunities


# In[11]:


#Transforming some numerical variables that are  categorical
filtered_df['Scenario Version'] = filtered_df['Scenario Version'].apply(str)


# ### Open_Opportunity for modeling, we will use Close_Opportunity for predicting point price

# In[12]:


Open_Opportunity = filtered_df.loc[filtered_df['Stage (group)'] != '5 Closed Disqualified, 5 Closed Lost, 5 Closed Won'].reset_index(drop=True)


# In[13]:


#filter the dataset with closed opportunities only
Close_Opportunity = filtered_df.loc[filtered_df['Stage (group)'] == '5 Closed Disqualified, 5 Closed Lost, 5 Closed Won'].reset_index(drop=True)


# In[14]:


print("\nThe train data size for Open_Opportunity is : {} ".format(Open_Opportunity.shape)) 
print("The predict data size for Close_Opportunity is : {} ".format(Close_Opportunity.shape))


# In[15]:


filtered_df.info() #working data brife info


# In[16]:


Open_Opportunity.dtypes.value_counts(normalize = True) #check the % of datatypes 


# In[17]:


#change the id data type to string
filtered_df[["Quote Id", "Scenario Id","Li Jpa" , "Contract Scenario Id"]] = filtered_df[["Quote Id", "Scenario Id","Li Jpa" , "Contract Scenario Id"]].astype(str) 


# In[18]:


filtered_df.describe().transpose().applymap("{0:,.2f}".format) #get a summary for numberical variables


# In[19]:


#get the descriptions for all categorical variables 
for i in filtered_df.select_dtypes(include=['object']).columns:
    print(i)
    print('Count of unique varuables:', len(filtered_df[i].unique()))
    print('Top 3 most common variables in ', i)
    print(filtered_df[i].value_counts().head(3),'\n')


# ### Feature extraction and categorize into 4 different data types

# In[20]:


categorical_data = ['Businessunit','Discount Level','Close Reason',
                       'Commitment Type','Contracted Flag','Drv Port Speed','Sp Port Speed', 'Feat Act Code Type','Feature Name',
                       'Nasp Type','New Order Flag','Next Order Flag','Opportunity Type','Pcm Touched','Product Name',
                       'Saleschannel','Scenario Category','Scenario Type','Scenario Version Status',
                       'Sp Acc Tech','Sp Cust Hand Off Type','Sp Dbw',
                       'Sp Vzb Lit Status','SP_IDS_PRICEPLAN', 'Stage (group)',
                       'Stage','Transaction Type','UNIQUE_PRODUCT_IDENTIFIER', 'Scenario Version',
                       'Unitdisc','Vertical Reporting Dsc']

numerical_data = ['Term',
                 'Commitment Amount',
                 'Li Bi Rec Direct Cost',
                 'Li Bi Rec List Price',
                 'Li Bi Rec Net Price',
                 'Price',
                 'Li Bi Rec Discount',
                 'Li Bi Rec Promo',
                 'Li Bi Rec Shared Cost',
                 'Li Bi Term Total Direct Cost',
                 'Li Bi Term Total Net Price',
                 'Li Bi Total Shared Cost',
                 'Li Dd Rec Net Price']

geo_data = ['City','State']

temporal_data = ['First Scenario Date']


# ### Determined 4 different of data types, we will encode variables in categorical_data

# In[21]:


#All variables for both Open_Opportunity and Closed_Opportunity
categorical_df = filtered_df[categorical_data]
numerical_df = filtered_df[numerical_data]
geo_df = filtered_df[geo_data]
temporal_df = filtered_df[temporal_data]
subset = pd.concat([categorical_df,numerical_df,geo_df,temporal_df], axis=1)


#All variables for Open_Opportunity only
categorical_df_open = Open_Opportunity[categorical_data]
numerical_df_open = Open_Opportunity[numerical_data]
geo_df_open = Open_Opportunity[geo_data]
temporal_df_open = Open_Opportunity[temporal_data]
subset_open = pd.concat([categorical_df_open,numerical_df_open,geo_df_open,temporal_df_open], axis=1)


#All variables for Closed_Opportunity only
categorical_df_close = Close_Opportunity[categorical_data]
numerical_df_close = Close_Opportunity[numerical_data]
geo_df_close = Close_Opportunity[geo_data]
temporal_df_close = Close_Opportunity[temporal_data]
subset_close = pd.concat([categorical_df_close,numerical_df_close,geo_df_close,temporal_df_close], axis=1)


print("\nThe train data size for subset is : {} ".format(subset.shape))
print("\nThe train data size for subset_open is : {} ".format(subset_open.shape)) 
print("\nThe predict data size for subset_close is : {} ".format(subset_close.shape))


# In[22]:


subset.loc[:,subset.dtypes==np.object].head(3)


# In[23]:


subset.loc[:,subset.dtypes==np.number].head(3)


# ### Data Processing

# ### Outliers Analysis and Removal for Colsed_Opportunity data which we will use for training and modeling, leave untouched for Open_Opportunity data here

# In[24]:


from scipy.stats import norm, skew
# Identifying Outliers with Skewness
skewed_feats = numerical_df_close.apply(lambda x: skew(x.dropna())).sort_values(ascending=False)
print("\nSkew in numerical features: \n")
skewness = pd.DataFrame({'Skew' :skewed_feats})
skewness


# In[25]:


g = sns.FacetGrid(pd.melt(subset_close[numerical_data]), col='variable',col_wrap=3,sharex=False, sharey=False)
g.map(sns.boxplot,'value')


# In[26]:


subset_close[numerical_data].describe()


# In[27]:


#We remove 2.5 percent percentile for both lower and upper boundary, identify these as outlier
Lower = subset_close.quantile(0.025)
Upper = subset_close.quantile(0.975)
IQR = Upper - Lower
print(IQR)


# In[28]:


Lower - 1.5 * IQR


# In[29]:


Upper + 1.5 * IQR


# In[30]:


print(subset_close.shape)
subset_close_final = subset_close[~((subset_close < (Lower - 1.5 * IQR)) |(subset_close > (Upper + 1.5 * IQR))).any(axis=1)]
print(subset_close_final.shape)


# **We lost about 17970 record from closed opportunity data, about 4% of data after applying this technqiue.**

# In[31]:


# count of zero value in each column prior to the log
for col in numerical_data:
    print(col, subset_close_final[subset_close_final[col]==0].shape[0])


# In[32]:


subset_close_final[numerical_data].describe() #Get a summary of numberical data for closed opportunity data after removing outliers


# Since Li Bi Rec Promo varaible is all 0s after outlier removing, we decide to remove this variable. 
# In addition, Li Bi Rec Discount has majority of 0s which is highly skewed with only influentical points, it will be removed as well. 

# #### Concat both Open_Opportunity and Closed_Opportunity after removing outliers from Closed_Opportunity data set, named subset_all_final

# In[33]:


subset_all_final = pd.concat([subset_close_final,subset_open])
#Drop varaible Li Bi Rec Promo since all 0 values
subset_all_final = subset_all_final.drop(['Li Bi Rec Promo','Li Bi Rec Discount'], axis=1)
print(subset_all_final.shape)


# In[34]:


#Redefine numerical_data column after droppping variable Li Bi Rec Promo
numerical_data  = ['Term',
                 'Commitment Amount',
                 'Li Bi Rec Direct Cost',
                 'Li Bi Rec List Price',
                 'Li Bi Rec Net Price',
                 'Price',
                 'Li Bi Rec Shared Cost',
                 'Li Bi Term Total Direct Cost',
                 'Li Bi Term Total Net Price',
                 'Li Bi Total Shared Cost',
                 'Li Dd Rec Net Price']


# #### For all data (combined closed and open opportunity data), we redefine four types of dataframe after moving the outliers from closed opportunity data

# In[35]:


categorical_df_all_final = subset_all_final[categorical_data]
numerical_df_all_final = subset_all_final[numerical_data]
geo_df_all_final = subset_all_final[geo_data]
temporal_df_all_final = subset_all_final[temporal_data]
print(categorical_df_all_final.shape,numerical_df_all_final.shape,geo_df_all_final.shape,temporal_df_all_final.shape)
#Output the data shape for closed_opportunity_final after removing the outliers


# ### Missing Value Imputation on all data (both Open_Opportunity and Closed_Opportunity) 

# In[36]:


#plot to show the missing data
msno.bar(subset_all_final)


# In[37]:


all_data_na = subset_all_final.isnull().sum() 
all_data_na = all_data_na.drop(all_data_na[all_data_na == 0].index).sort_values(ascending=False)
missing_data = pd.DataFrame({'Missing Value Count' :all_data_na})
missing_data #Give us a list of missing value count 


# In[38]:


all_data_na_ratio = (subset_all_final.isnull().sum() / len(subset)) * 100
all_data_na_ratio = all_data_na_ratio.drop(all_data_na_ratio[all_data_na_ratio == 0].index).sort_values(ascending=False)
missing_data_ratio = pd.DataFrame({'Missing Ratio' :all_data_na_ratio})
missing_data_ratio #Give us a list of missing value ratio 


# In[39]:


#need to imputate numeric_data with median for numerical_df_all_final
numerical_df_all_final.fillna(numerical_df_all_final.median(), inplace=True)


# In[40]:


#need to imputate categorical_data with Unknown for categorical_df_all_final
categorical_df_all_final = categorical_df_all_final.fillna("Unknown")


# In[41]:


#no need to imputate geo_df_all_final
geo_df_all_final.isnull().sum()


# In[42]:


#need to imputate temporal data with "01/01/1900" for temporal_df_all_final
temporal_df_all_final = temporal_df_all_final.fillna("01/01/1900")


# #### For all data (combined closed and open opportunity data), we concatenate  all four types data dataframe after the imputation, named subset_all_final_com

# In[43]:


#Concat all four types of variable 
subset_all_final_com = pd.concat([numerical_df_all_final,categorical_df_all_final,geo_df_all_final,temporal_df_all_final], axis=1)
subset_all_final_com.shape


# In[44]:


all_data_na2 = subset_all_final_com.isnull().sum() 
all_data_na2 = all_data_na2.drop(all_data_na2[all_data_na2 == 0].index).sort_values(ascending=False)
missing_data2 = pd.DataFrame({'Missing Value Count' :all_data_na2})
missing_data2 #check the null value again, there is no missing value now


# There is no missing value after the imputation.

# #### Resplit open and close opportunity data after imputing the missing value

# In[45]:


#define Open_Opportunity_final for the open opportunity and Close_Opportunity_final for the closed opportunity.
Open_Opportunity_final = subset_all_final_com.loc[subset_all_final_com['Stage (group)'] != '5 Closed Disqualified, 5 Closed Lost, 5 Closed Won'].reset_index(drop=True)
Close_Opportunity_final = subset_all_final_com.loc[subset_all_final_com['Stage (group)'] == '5 Closed Disqualified, 5 Closed Lost, 5 Closed Won'].reset_index(drop=True)
print('Open_Opportunity_final: ', Open_Opportunity_final.shape)
print('Close_Opportunity_final: ', Close_Opportunity_final.shape)


# Now we can see we have 336004 records for open opportunity and 423111 for closed 423111 opportunity, both has 45 columns.

# In[46]:


#All variables for Open_Opportunity only, we redefine four types of data after moving the outliers and missing value imputation
categorical_df_open_final = Open_Opportunity_final[categorical_data]
numerical_df_open_final = Open_Opportunity_final[numerical_data]
geo_df_open_final = Open_Opportunity_final[geo_data]
temporal_df_open_final = Open_Opportunity_final[temporal_data]
print(categorical_df_open_final.shape,numerical_df_open_final.shape,geo_df_open_final.shape,temporal_df_open_final.shape)


# **Above is the processed Open_Opportunity data we will use for the pricing prediction, data shapes displayed above.**

# In[47]:


#All variables for Closed_Opportunity only, we redefine four types of data after moving the outliers and missing value imputation
categorical_df_close_final = Close_Opportunity_final[categorical_data]
numerical_df_close_final = Close_Opportunity_final[numerical_data]
geo_df_close_final = Close_Opportunity_final[geo_data]
temporal_df_close_final = Close_Opportunity_final[temporal_data]
print(categorical_df_close_final.shape,numerical_df_close_final.shape,geo_df_close_final.shape,temporal_df_close_final.shape)


# **Above is the processed Closed_Opportunity data we will use for the futuer modeling, data shapes displayed above.**

# ### Initial Skewness Analysis

# In[48]:


# Find skewed numerical features
skew_features = numerical_df_close_final.apply(lambda x: skew(x)).sort_values(ascending=False)

high_skew = skew_features[skew_features > 0.5]
skew_index = high_skew.index

print("There are {} numerical features with Skew > 0.5 :".format(high_skew.shape[0]))
skewness = pd.DataFrame({'Skew' :high_skew})
skew_features


# In[49]:


#Analysis for original target variable
fig,(ax1,ax2,ax3) = plt.subplots(1,3,constrained_layout=True,figsize=(10,3))

y1 = numerical_df_close_final.Price
sns.boxplot(y = y1, ax=ax1)
ax1.set_title("Boxplot of Price")
sns.distplot(y1, ax=ax2,kde=False, fit=stats.norm)
ax2.set_title("Histrogram of Price")
stats.probplot(y1,plot=ax3)
ax3.set_title("QQplot of Price")
print("Kurtosis: %f" % numerical_df_close_final.Price.kurt())
print("Skewness: %f" % numerical_df_close_final.Price.skew())


# ## Feature Engineering on both open and closed opportunity data
# ### Log Transformation
# ### StandardScaler

# ### Apply log transformation

# In[50]:


#define the original numberical columns including target variable Price 
original_num_columns = numerical_df_close_final.columns

#define the original numberical columns excluding target variable Price 
original_num_columns1 = numerical_df_close_final.drop(['Price'], axis=1).columns


# In[51]:


#converting each numberical columns into its log variation for open opportunity 
for column in original_num_columns1:
        numerical_df_open_final[column + '_log'] = np.log1p(numerical_df_open_final[column])
log_columns = numerical_df_open_final.columns[numerical_df_open_final.columns.str.contains('_log')].tolist()
# getting a quick pick at the skewed data
numerical_df_open_final.head(2)


# In[52]:


#converting each numberical into its log variation for closed opportunity 
for column in original_num_columns1:
        numerical_df_close_final[column + '_log'] = np.log1p(numerical_df_close_final[column])
log_columns = numerical_df_close_final.columns[numerical_df_close_final.columns.str.contains('_log')].tolist()
# getting a quick pick at the skewed data
numerical_df_close_final.head(2)


# **After applying the log transformation, there might be some small possibilities na value occurs, check the null values again.**

# In[53]:


numerical_df_open_final.isnull().sum()[numerical_df_open_final.isnull().sum()>0]


# In[54]:


numerical_df_close_final.isnull().sum()[numerical_df_close_final.isnull().sum()>0]


# In[55]:


#Inpute nan value with 0 after log trnasformation
numerical_df_open_final = numerical_df_open_final.fillna(0)
numerical_df_close_final = numerical_df_close_final.fillna(0)


# ### Apply StandardScaler for both closed and open opportunity based on log transformed data

# In[56]:


# Standardizing open opportunity data
scaler1 = StandardScaler()
numerical_scaled = scaler1.fit_transform(numerical_df_open_final[log_columns].values)
new_col = [l + "_ss" for l in log_columns]
numerical_df_open_final_scaled = pd.DataFrame(numerical_scaled, columns=new_col)
#add scaled data to our numerical_df_open_final data
numerical_df_open_final = pd.concat([numerical_df_open_final, numerical_df_open_final_scaled], axis=1)
pd.set_option("display.max_columns", numerical_df_open_final.shape[1]) 
# Log transformation on target variable 
numerical_df_open_final['Price_log'] = np.log1p(numerical_df_open_final.Price)
numerical_df_open_final.head(2)


# In[57]:


# Standardizing closed opportunity data
scaler2 = StandardScaler()
numerical_scaled = scaler2.fit_transform(numerical_df_close_final[log_columns].values)
new_col = [l + "_ss" for l in log_columns]
numerical_df_close_final_scaled = pd.DataFrame(numerical_scaled, columns=new_col)
#add scaled data to our numerical_df_close_final data
numerical_df_close_final = pd.concat([numerical_df_close_final, numerical_df_close_final_scaled], axis=1)
pd.set_option("display.max_columns", numerical_df_close_final.shape[1]) 
# Log transformation on target variable 
numerical_df_close_final['Price_log'] = np.log1p(numerical_df_close_final.Price)
numerical_df_close_final.head(2)


# In[58]:


#only for original open data columns
numerical_df_open_final[original_num_columns].head(2)


# In[59]:


#only for original closed data columns
numerical_df_close_final[original_num_columns].head(2)


# In[60]:


#Create a shortlist for the modeling named new_col
new_col.append('Price_log')
new_col


# In[61]:


#only for scaled open data columns
numerical_df_open_final[new_col].head(2)


# In[62]:


#only for scaled closed data columns
numerical_df_close_final[new_col].head(2)


# #### Below workdata_open is the unscaled data for the final OPEN opportunity data

# In[63]:


workdata_open = pd.concat([categorical_df_open_final,numerical_df_open_final[original_num_columns],geo_df_open_final,temporal_df_open_final], axis=1)
workdata_open.shape


# #### Below workdata is the unscaled data for the final CLOSED opportunity data we will be used for further EDA

# In[64]:


#this is the unscaled data for closed opportunity
workdata = pd.concat([categorical_df_close_final,numerical_df_close_final[original_num_columns],geo_df_close_final,temporal_df_close_final], axis=1)
workdata.shape


# #### Below workdata_open_scaled is the scaled data for the final OPEN opportunity data we will be used for the prediction 

# In[65]:


#this is the logscaled open opportunity data which we will be using for the predicting
workdata_open_scaled = pd.concat([categorical_df_open_final,numerical_df_open_final[new_col],geo_df_open_final,temporal_df_open_final], axis=1)
workdata_open_scaled.shape


# #### Below workdata_scaled is the scaled data for the final CLOSED opportunity data we will be used for the modeling 

# In[66]:


#this is the logscaled closed data which we will be using in the modeling
workdata_scaled = pd.concat([categorical_df_close_final,numerical_df_close_final[new_col],geo_df_close_final,temporal_df_close_final], axis=1)
workdata_scaled.shape


# In[67]:


workdata_scaled.info() #check the workdata_scaled info again before modeling 


# ### EDA on original closed(training) dataset - workdata

# In[68]:


#scatterplot to magnify and look at the selected variable
sns.set()
sns.pairplot(workdata[original_num_columns])
plt.show()


# In[69]:


#EDA with respect to target variable for workdataset
g = sns.factorplot(x="Discount Level", y="Price", col="Product Name",
                    data=workdata,kind="bar", ci=None, size=7, aspect=2)
plt.show()


# In[70]:


g1 = sns.FacetGrid(workdata, hue = 'Transaction Type', aspect=5,size=5 )
g1.map(sns.kdeplot, 'Term', shade= True )
g1.set(xlim=(0 , workdata['Term'].max()))
g1.add_legend()
plt.show()


# In[71]:


pd.crosstab(workdata['Discount Level'],workdata['Stage (group)']).style.background_gradient(cmap='PuBu')


# In[72]:


# crosstab to get some summaries from subset2
pd.crosstab(workdata['Discount Level'],[workdata['Product Name'],workdata['Transaction Type']]).style.background_gradient(cmap='PuBu')


# In[73]:


pd.crosstab(workdata['Drv Port Speed'],workdata['Scenario Type']).style.background_gradient(cmap='PuBu')


# In[74]:


pd.crosstab(workdata['Sp Port Speed'],workdata['Scenario Version Status']).style.background_gradient(cmap='PuBu')


# In[75]:


pd.crosstab(workdata['Nasp Type'],[workdata['Transaction Type'],workdata['Unitdisc']], rownames=['Nasp Type']).style.background_gradient(cmap='PuBu')


# ### Skewness analysis for colsed_opportunity data after standard scaling use workdata_scaled dataframe

# In[76]:


#Looking at the data after cleaning the outliers, applying the standard scaling and logtransformation
chart1 = sns.lmplot(x='Term_log_ss', y='Price_log', data=workdata_scaled, scatter_kws={'alpha':0.75})
chart2 = sns.lmplot(x='Li Bi Rec Direct Cost_log_ss', y='Price_log', data=workdata_scaled, scatter_kws={'alpha':0.75})
chart3 = sns.lmplot(x='Li Bi Rec List Price_log_ss', y='Price_log', data=workdata_scaled, scatter_kws={'alpha':0.75})
chart4 = sns.lmplot(x='Li Bi Rec Net Price_log_ss', y='Price_log', data=workdata_scaled, scatter_kws={'alpha':0.75})


# In[77]:


# Find skewed numerical features
skew_features = workdata_scaled[new_col].apply(lambda x: skew(x)).sort_values(ascending=False)

high_skew = skew_features[skew_features > 0.5]
skew_index = high_skew.index

print("There are {} numerical features with Skew > 0.5 :".format(high_skew.shape[0]))
skewness = pd.DataFrame({'Skew' :high_skew})
skew_features


# In[78]:


#Analysis again for original target variable after the feature engineering 
fig,(ax1,ax2,ax3) = plt.subplots(1,3,constrained_layout=True,figsize=(10,3))

y1 = workdata_scaled.Price_log
sns.boxplot(y = y1, ax=ax1)
ax1.set_title("Boxplot of Price_log")
sns.distplot(y1, ax=ax2,kde=False, fit=stats.norm)
ax2.set_title("Histrogram of Price_log")
stats.probplot(y1,plot=ax3)
ax3.set_title("QQplot of Price_log")
print("Kurtosis: %f" % workdata_scaled.Price_log.kurt())
print("Skewness: %f" % workdata_scaled.Price_log.skew())


# In[79]:


# Looking at the standard scaled version of our variables
fig, axes = plt.subplots(nrows = 3, ncols = 3)
axes = axes.flatten()
fig.set_size_inches(15, 15)

for ax, col in zip(axes, new_col):
    sns.distplot(workdata_scaled[col], ax = ax)
    ax.set_title(col)


# #### After perfroming the data imputation, log transformation, and standard scaler and skewness analysis on the numberical data, workdata_scaled is the one we will use for modeling, workdata_open_scaled is the one we will use for prediction. Look at the distribution plots, skewness and statistical plots, we can see our dataset workdata_scaled is in good shape for the modeling.

# In[80]:


workdata_scaled.describe()


# In[81]:


workdata_open_scaled.describe()


# ### Checking correlation matrix using heatmap to find features influencing the target variable "Price" the most.

# In[82]:


sns.heatmap(workdata_scaled.corr(),square=True, vmax = 1, vmin = 0.3, xticklabels=True, yticklabels=True, cmap='YlGn') 
fig=plt.gcf()
fig.set_size_inches(12,9)
plt.show()


# In[83]:


plt.figure(figsize=(18,12))
sns.heatmap((workdata_scaled.corr()),annot=True)
plt.title('Correlation Matrix', fontsize = 25)
plt.show()


# ### Use OneHot encoder for categorical variables for CLOSED opportunity, define workdata_scaled_digit and we will use workdata_scaled_digit dataframe for modeling.

# In[84]:


OneHotencoder = OneHotEncoder()
OneHotencoder = ce.OneHotEncoder(cols=categorical_data,handle_unknown='ignore',return_df=True,use_cat_names=True)
workdata_scaled_digit = OneHotencoder.fit_transform(workdata_scaled)


# ### Use OneHot encoder for categorical variables for OPEN opportunity, define testdata_scaled_digit and we will use testdata_scaled_digit dataframe for prediction.

# In[85]:


OneHotencoder = OneHotEncoder()
OneHotencoder = ce.OneHotEncoder(cols=categorical_data,handle_unknown='ignore',return_df=True,use_cat_names=True)
testdata_scaled_digit = OneHotencoder.fit_transform(workdata_open_scaled)


# ### Using Label encodering for geo_data and temporal_data that may contain information in ordering set, applied to workdata_scaled_digit dataset (CLOSED opportunity).

# In[86]:


from sklearn.preprocessing import LabelEncoder

cols = ['City','State','First Scenario Date']

for c in cols:
    lbl = LabelEncoder() 
    lbl.fit(list(workdata_scaled_digit[c].values)) 
    workdata_scaled_digit[c] = lbl.transform(list(workdata_scaled_digit[c].values))

# shape        
print('Shape all_closed_data: {}'.format(workdata_scaled_digit.shape))


# ### Using Label encodering for geo_data and temporal_data that may contain information in ordering set, applied to testdata_scaled_digit (OPEN opportunity).

# In[87]:


for c in cols:
    lbl = LabelEncoder() 
    lbl.fit(list(testdata_scaled_digit[c].values)) 
    testdata_scaled_digit[c] = lbl.transform(list(testdata_scaled_digit[c].values))

# shape        
print('Shape all_open_data: {}'.format(testdata_scaled_digit.shape))


# ### Check and remove the dummy variables that either not in workdata_scaled_digit or not in testdata_scaled_digit

# In[88]:


# Get missing columns, exist in workdata_scaled_digit but not in testdata_scaled_digit dataset
missing_cols1 = set( workdata_scaled_digit.columns ) - set( testdata_scaled_digit.columns )
missing_cols1 = missing_cols1- set(workdata_scaled.columns)


# In[89]:


# Get missing columns, exist in testdata_scaled_digit but not in workdata_scaled_digit dataset
missing_cols2 = set( testdata_scaled_digit.columns ) - set( workdata_scaled_digit.columns ) 
missing_cols2 = missing_cols2 - set(workdata.columns)


# In[90]:


# We exlucde missing_cols1 and missing_cols2 from workdata_scaled_digit and testdata_scaled_digit dataset
Col_to_exlcude = missing_cols1.union(missing_cols2)
list(Col_to_exlcude)


# In[91]:


#Remove variable in testdata_scaled_digit but not in workdata_scaled_digit
cols1 = [col for col in workdata_scaled_digit.columns if col not in Col_to_exlcude]
workdata_scaled_digit = workdata_scaled_digit[cols1]
print('Shape all_closed_data: {}'.format(workdata_scaled_digit.shape))


# In[92]:


#Remove variable in workdata_scaled_digit but not in testdata_scaled_digit
cols2 = [col for col in testdata_scaled_digit.columns if col not in Col_to_exlcude]
testdata_scaled_digit = testdata_scaled_digit[cols2]
print('Shape all_open_data: {}'.format(testdata_scaled_digit.shape))


# ### Data Split into train and validation

# In[93]:


#Define predictive dataset with dropping the predictive variable
Pred_data = testdata_scaled_digit.drop(['Price_log'],axis=1)
Pred_data.shape


# In[94]:


#Split the train and test data from workdata_scaled_digit dataset
y = workdata_scaled_digit.Price_log
X = workdata_scaled_digit.drop(['Price_log'],axis=1)
from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(X,y,                                         test_size = 0.2, random_state=42)
X_train.shape, X_test.shape, y_train.shape, y_test.shape


# In[95]:


#Create five folds for the validation 
from sklearn.model_selection import GridSearchCV, train_test_split, KFold

kfold =KFold(n_splits=5, shuffle=True, random_state=42)
cnt = 1
# split()  method generate indices to split data into training and test set.
for train_index, test_index in kfold.split(X, y):
    print(f'Fold:{cnt}, Train set: {len(train_index)}    , Test set:{len(test_index)}')
    cnt += 1


# ### Regression 

# In[96]:


# Importing sklearn libraries
from sklearn.linear_model import Ridge, RidgeCV, ElasticNet, Lasso, LassoCV, LassoLarsCV, LinearRegression
from sklearn.model_selection import cross_val_score
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import RobustScaler

#from sklearn.linear_model import ElasticNet, Lasso, Ridge, Lasso regressions as well as metrics
from sklearn.ensemble import RandomForestRegressor,  GradientBoostingRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.base import BaseEstimator, TransformerMixin, RegressorMixin, clone
from sklearn.model_selection import KFold, cross_val_score, train_test_split
from sklearn.metrics import mean_squared_error

from yellowbrick.regressor import ResidualsPlot
from yellowbrick.model_selection import LearningCurve
from yellowbrick.regressor import PredictionError


# In[97]:


#Define helper functions that used in the modeling
def cross_val(model):
    cv_rmse = np.sqrt(-cross_val_score(model, X, y,     scoring="neg_mean_squared_error", cv=kfold))
    model_score = model.score(X,y)
    return cv_rmse.mean(), cv_rmse.std(), model_score
def evaluate_model(true, predicted):
    mse = metrics.mean_squared_error(true, predicted)
    rmse = np.sqrt(metrics.mean_squared_error(true, predicted))
    R2_squared = metrics.r2_score(true, predicted)
    return mse,rmse, R2_squared
def print_evaluate_results(true, predicted):
    mse = metrics.mean_squared_error(true, predicted)
    rmse = np.sqrt(metrics.mean_squared_error(true, predicted))
    R2_squared = metrics.r2_score(true, predicted)
    print('MSE:', round(mse,5), 'RMSE:', round(rmse,5),           'R2_Squared', round(R2_squared,5))
def print_cross_val(model):
    cv_rmse = np.sqrt(-cross_val_score(model, X, y,     scoring="neg_mean_squared_error", cv=kfold))
    model_score = model.score(X,y)
    print('CV Results:', 'Mean:',round(cv_rmse.mean(),5),'Std:',round(cv_rmse.std(),5), 'Model R2_Squared', round(model_score,5))
    
def plot_learning_curves(model, X, y):
    X_train, X_val, y_train, y_val =     train_test_split(X, y, test_size=0.33, random_state=10)
    train_errors, val_errors = [], []
    for m in range(1, len(X_train)):
        model.fit(X_train[:m], y_train[:m])
        y_train_predict = model.predict(X_train[:m])
        y_val_predict = model.predict(X_val)
        train_errors.append(mean_squared_error                            (y_train[:m], y_train_predict))
        val_errors.append(mean_squared_error(y_val, y_val_predict))
    plt.plot(np.sqrt(train_errors), "r-+", linewidth=2, label="train")
    plt.plot(np.sqrt(val_errors), "b-", linewidth=3, label="val")
    plt.legend(loc="upper right", fontsize=14) 
    plt.xlabel("Training set size", fontsize=14) 
    plt.ylabel("RMSE", fontsize=14)


# ### Linear Regression

# In[98]:


#Fit linear regression model
lin_reg = LinearRegression()
lin_reg.fit(X_train,y_train)
print(lin_reg.intercept_)


# In[99]:


coeff_df = pd.DataFrame(lin_reg.coef_, X.columns, columns=['Coefficient'])
coeff_df.sort_values(['Coefficient'], ascending=False)


# In[100]:


en_coefs = pd.Series(lin_reg.coef_, index = X_train.columns)
imp_coefs = pd.concat([en_coefs.sort_values().head(),(en_coefs[en_coefs==0]),en_coefs.sort_values().tail()])
imp_coefs.plot(kind = "barh")
plt.show()


# In[101]:


y_pred_train=lin_reg.predict(X_train)
y_pred_test=lin_reg.predict(X_test)


# In[102]:


print('Test set evaluation:\n')
print_evaluate_results(y_test,y_pred_test)
print('======')
print('Train set evaluation:\n')
print_evaluate_results(y_train,y_pred_train)
print('======')
print('CV model evaluation:\n')
print_cross_val(lin_reg)


# In[103]:


#Get the statistics model summary 
model1 = pd.DataFrame(data = [['Linear',*evaluate_model(y_test,y_pred_test), cross_val(lin_reg)[0],cross_val(lin_reg)[1],                cross_val(lin_reg)[2]]],
    columns=['Model','MSE','RMSE', 'R Squared', \
    'CV MSE mean', 'CV MSE std', 'Model score'])
print(model1)


# In[104]:


#ElasticNet Regression Plots
fig, ax = plt.subplots(ncols=3, figsize=(16,4))
visualizer1 = PredictionError(lin_reg,ax=ax[0])
visualizer1.fit(X_train, y_train)
visualizer1.score(X_test, y_test)
visualizer1.finalize()
visualizer2 = ResidualsPlot(lin_reg,ax=ax[1])
visualizer2.fit(X_train, y_train)
visualizer2.score(X_test, y_test)
visualizer2.finalize()
visualizer3 = LearningCurve(lin_reg,ax=ax[2])
visualizer3.fit(X_train, y_train)
visualizer3.finalize()


# In[105]:


reg_pred = lin_reg.predict(Pred_data)
result1 = pd.DataFrame()
# exp(x) - 1, the inverse of log1p
result1['Point_Price1'] = np.exp(reg_pred)-1
#Output the result from the linear model 
result1


# In[106]:


result1.describe()


# ### Ridge Regression

# In[107]:


#Fit Ridge regression model and turn the model to 
#obtain best alpha value and the best model score
ridge=Ridge()
params= {'alpha': [0.0005,0.001,0.1, 1, 10]}
ridge_grid=GridSearchCV(ridge, param_grid=params,cv=kfold,scoring='neg_mean_squared_error', n_jobs=-1)
ridge_grid.fit(X_train,y_train)
alpha = ridge_grid.best_params_
ridge_score = ridge_grid.best_score_
ridge_reg=Ridge(alpha=alpha['alpha'])
ridge_reg.fit(X_train,y_train)
y_predrid_train=ridge_reg.predict(X_train)
y_predrid_test=ridge_reg.predict(X_test)
print("The best alpha value is:",alpha['alpha'],'with score:',ridge_score)


# In[108]:


coeff_df = pd.DataFrame(ridge_reg.coef_, X.columns, columns=['Coefficient'])
coeff_df.sort_values(['Coefficient'], ascending=False)


# In[109]:


en_coefs = pd.Series(ridge_reg.coef_, index = X_train.columns)
imp_coefs = pd.concat([en_coefs.sort_values().head(),(en_coefs[en_coefs==0]),en_coefs.sort_values().tail()])
imp_coefs.plot(kind = "barh")
plt.show()


# In[110]:


ridge_predss = pd.Series(y_predrid_test)
ridge_predss


# In[111]:


#To get a list of alpha with corresponding mean test scores 
res = pd.DataFrame(ridge_grid.cv_results_['params'])
res['score'] = ridge_grid.cv_results_['mean_test_score']
res.sort_values('score',ascending=False).head(3)


# In[112]:


print('Test set evaluation:\n')
print_evaluate_results(y_test,y_predrid_test)
print('======')
print('Train set evaluation:\n')
print_evaluate_results(y_train,y_predrid_train)
print('======')
print('CV model evaluation:\n')
print_cross_val(ridge_reg)


# In[113]:


#Get the statistics model summary 
model2 = pd.DataFrame(data = [['Ridge',*evaluate_model(y_test,y_predrid_test), cross_val(ridge_reg)[0],cross_val(ridge_reg)[1],                cross_val(ridge_reg)[2]]],
    columns=['Model','MSE','RMSE', 'R Squared', \
    'CV MSE mean', 'CV MSE std', 'Model score'])
print(model2)


# In[114]:


#Ridge Regression Analysis Plots
fig, ax = plt.subplots(ncols=3, figsize=(16,4))
visualizer1 = PredictionError(ridge_reg,ax=ax[0])
visualizer1.fit(X_train, y_train)
visualizer1.score(X_test, y_test)
visualizer1.finalize()
visualizer2 = ResidualsPlot(ridge_reg,ax=ax[1])
visualizer2.fit(X_train, y_train)
visualizer2.score(X_test, y_test)
visualizer2.finalize()
visualizer3 = LearningCurve(ridge_reg,ax=ax[2])
visualizer3.fit(X_train, y_train)
visualizer3.finalize()


# In[115]:


rid_pred = ridge_reg.predict(Pred_data)
result2 = pd.DataFrame()

# exp(x) - 1, the inverse of log1p
result2['Point_Price2'] = np.exp(rid_pred)-1
#Output the result from the ridge model 
result2


# ### Lasso Regression

# In[116]:


#Fit Lasso regression model and turn the model to 
#obtain best alpha value and the best model score
lasso_reg =Lasso()
params= {'alpha': [0.001,0.1,1]}
lasso_grid = GridSearchCV(lasso_reg, param_grid=params,cv=kfold,scoring='neg_mean_squared_error', n_jobs=-1)
lasso_grid.fit(X_train,y_train)
alpha = lasso_grid.best_params_
lasso_score = lasso_grid.best_score_
lasso_reg=Lasso(alpha=alpha['alpha'])
lasso_reg.fit(X_train,y_train)
y_predlas_train=lasso_reg.predict(X_train)
y_predlas_test=lasso_reg.predict(X_test)
print("The best alpha value is:",alpha['alpha'],'with score:',lasso_score)


# In[117]:


coeff_df = pd.DataFrame(lasso_reg.coef_, X.columns, columns=['Coefficient'])
coeff_df.sort_values(['Coefficient'], ascending=False)


# In[118]:


en_coefs = pd.Series(lasso_reg.coef_, index = X_train.columns)
imp_coefs = pd.concat([en_coefs.sort_values().head(1),(en_coefs[en_coefs==0]),en_coefs.sort_values().tail(1)])
imp_coefs.plot(kind = "barh")
plt.show()


# In[119]:


lasso_predss = pd.Series(y_predlas_test)
lasso_predss


# In[120]:


#To get a list of alpha with corresponding mean test scores 
res = pd.DataFrame(lasso_grid.cv_results_['params'])
res['score'] = lasso_grid.cv_results_['mean_test_score']
res.sort_values('score',ascending=False).head(3)


# In[121]:


print('Test set evaluation:\n')
print_evaluate_results(y_test,y_predlas_test)
print('======')
print('Train set evaluation:\n')
print_evaluate_results(y_train,y_predlas_train)
print('======')
print('CV model evaluation:\n')
print_cross_val(lasso_reg)


# In[122]:


#Get the statistics model summary 
model3 = pd.DataFrame(data = [['Lasso',*evaluate_model(y_test,y_predlas_test), cross_val(lasso_reg)[0],cross_val(lasso_reg)[1],cross_val(lasso_reg)[2]]],columns=['Model','MSE','RMSE', 'R Squared','CV MSE mean', 'CV MSE std', 'Model score'])
print(model3)


# In[123]:


#Lasso Regression Analysis Plots
fig, ax = plt.subplots(ncols=3, figsize=(16,4))
visualizer1 = PredictionError(lasso_reg,ax=ax[0])
visualizer1.fit(X_train, y_train)
visualizer1.score(X_test, y_test)
visualizer1.finalize()
visualizer2 = ResidualsPlot(lasso_reg,ax=ax[1])
visualizer2.fit(X_train, y_train)
visualizer2.score(X_test, y_test)
visualizer2.finalize()
visualizer3 = LearningCurve(lasso_reg,ax=ax[2])
visualizer3.fit(X_train, y_train)
visualizer3.finalize()


# In[124]:


laso_pred = lasso_reg.predict(Pred_data)
result3 = pd.DataFrame()

# exp(x) - 1, the inverse of log1p
result3['Point_Price3'] = np.exp(laso_pred)-1
#Output the result from the lasso model 
result3


# ### ElasticNet Regression

# In[125]:


#Fit elasticnet regression model and turn the model to 
#obtain best estimators and the best model score
elasticnet_reg =ElasticNet()
params= {'alpha': [0.001, 0.01,1],
        'l1_ratio': [0.05,0.5,1]}
elasticnet_grid=GridSearchCV(elasticnet_reg, param_grid=params,cv=kfold,scoring='neg_mean_squared_error', n_jobs=-1)
elasticnet_grid.fit(X_train,y_train)
elasticnet_reg = elasticnet_grid.best_estimator_
elasticnet_score = elasticnet_grid.best_score_
elasticnet_reg.fit(X_train,y_train)
y_predent_train=elasticnet_reg.predict(X_train)
y_predent_test=elasticnet_reg.predict(X_test)
print("The best estimator is:",elasticnet_reg,       'with score:',elasticnet_score)


# In[126]:


coeff_df = pd.DataFrame(elasticnet_reg.coef_, X.columns, columns=['Coefficient'])
coeff_df.sort_values(['Coefficient'], ascending=False)


# In[127]:


en_coefs = pd.Series(elasticnet_reg.coef_, index = X_train.columns)
imp_coefs = pd.concat([en_coefs.sort_values().head(10),(en_coefs[en_coefs==0]),en_coefs.sort_values().tail(10)])
imp_coefs.plot(kind = "barh")
plt.show()


# In[128]:


#To get a list of alpha with corresponding mean test scores 
res = pd.DataFrame(elasticnet_grid.cv_results_['params'])
res['score'] = elasticnet_grid.cv_results_['mean_test_score']
res.sort_values('score',ascending=False).head(3)


# In[129]:


print('Test set evaluation:\n')
print_evaluate_results(y_test,y_predent_test)
print('======')
print('Train set evaluation:\n')
print_evaluate_results(y_train,y_predent_train)
print('======')
print('CV model evaluation:\n')
print_cross_val(elasticnet_reg)


# In[130]:


#Get the statistics model summary 
model4 = pd.DataFrame(data = [['ElasticNet',*evaluate_model(y_test,y_predent_test), cross_val(elasticnet_reg)[0],cross_val(elasticnet_reg)[1],cross_val(elasticnet_reg)[2]]],
columns=['Model','MSE','RMSE', 'R Squared', \
'CV MSE mean', 'CV MSE std', 'Model score'])
print(model4)


# In[131]:


#ElasticNet Regression Plots
fig, ax = plt.subplots(ncols=3, figsize=(16,4))
visualizer1 = PredictionError(elasticnet_reg,ax=ax[0])
visualizer1.fit(X_train, y_train)
visualizer1.score(X_test, y_test)
visualizer1.finalize()
visualizer2 = ResidualsPlot(elasticnet_reg,ax=ax[1])
visualizer2.fit(X_train, y_train)
visualizer2.score(X_test, y_test)
visualizer2.finalize()
visualizer3 = LearningCurve(elasticnet_reg,ax=ax[2])
visualizer3.fit(X_train, y_train)
visualizer3.finalize()


# In[132]:


Elas_pred = elasticnet_reg.predict(Pred_data)
result4 = pd.DataFrame()

# exp(x) - 1, the inverse of log1p
result4['Point_Price4'] = np.exp(Elas_pred)-1
#Output the result from the elasticnet model 
result4


# ### Convolutional Neural Network

# In[133]:


import keras
from keras.models import Sequential
from keras.layers import Dense, Dropout, Activation, Flatten
from keras.layers import Conv2D, MaxPooling2D

from keras import backend as K
from keras.models import Sequential
from keras.layers import Dense, Conv1D, Flatten
from sklearn.metrics import mean_squared_error
import matplotlib.pyplot as plt

from tensorflow.keras.layers import BatchNormalization


# In[134]:


X_train.shape


# In[135]:


from keras.models import Sequential


# In[136]:


# Model Build

model = Sequential()
model.add(Conv1D(32, 2, activation="relu", input_shape=(475, 1)))
model.add(Flatten())
model.add(Dense(64, activation="relu"))
model.add(Dense(1))
model.compile(loss="mse", optimizer="adam")

model.summary()


# In[137]:


model.fit(X_train, y_train, batch_size=200,epochs=5, verbose=0)


# In[138]:


score = model.evaluate(X_test, y_test, verbose=0)
print('Test loss:', score)


# In[139]:


y_predcnn_train=model.predict(X_train)
y_predcnn_test=model.predict(X_test)

print('Test set evaluation:\n')
print_evaluate_results(y_test,y_predcnn_test)
print('======')
print('Train set evaluation:\n')
print_evaluate_results(y_train,y_predcnn_train)


# In[140]:


cnn_pred_test = model.predict_generator(X_test)


# In[141]:


cnn_pred_test = pd.Series(cnn_pred_test.flatten())


# In[142]:


y_test = y_test.reset_index(drop=True)


# In[143]:


#Lasso Regression Model Actual Vs. Predicted Plot
fig, ax = plt.subplots(ncols=3, figsize=(14,4))
ax[0].scatter(y_test, cnn_pred_test,alpha=.75,               color='b', label = "Validation data")
ax[0].plot([-5, 5], [-5, 5], '--k', color='r')
ax[0].set_ylabel('Predicted Prices');
ax[0].set_xlabel('Actual Prices');
ax[0].set_title("Lasso Regression Actual vs. Predicted Prices")
ax[0].legend(loc = "upper left")
#Lasso Regression Model Residual Plot Against Predicted Value
resid1 = cnn_pred_test - y_test
ax[1].scatter(y_predcnn_test, resid1, c = "b",               label = "Validation data")
ax[1].set_title("CNN Residuals vs. Predicted Value" )
ax[1].set_xlabel("Predicted values")
ax[1].set_ylabel("Residuals")
ax[1].legend(loc = "upper left")
ax[1].hlines(y = 0, xmin = -5, xmax = 5,             linestyles='dashed', color = "red")
residuals = y_test - cnn_pred_test
sns.distplot(residuals)
ax[2].set_title("Residuals Distribution" )
plt.show()


# In[144]:


#Get the statistics model summary 
model5 = pd.DataFrame(data = [['CNN',*evaluate_model(y_test,cnn_pred_test),]],
columns=['Model','MSE','RMSE', 'R Squared'])
print(model5)


# In[145]:


cnn_pred = model.predict(Pred_data)
cnn_pred = pd.Series(cnn_pred.flatten())
result5 = pd.DataFrame()

# exp(x) - 1, the inverse of log1p
result5['Point_Price5'] = np.exp(cnn_pred)-1
#Output the result from the cnn model 
result5


# In[146]:


#Define the validation data from original open opportunity point price 
Val_data = pd.DataFrame(np.exp(testdata_scaled_digit['Price_log'])-1)
Val_data


# In[147]:


final_result = pd.concat([Val_data,result1, result2,result3, result4,result5], axis=1, join="inner")


# In[148]:


final_result.describe().round(1)


# In[149]:


final_result.to_csv('final_result.csv') #save the prediction results to excel


# ### Model Evalutation and Comparsion 

# In[150]:


frames = [model1,model2, model3, model4, model5]
result = pd.concat(frames)
result.set_index("Model", drop = True, inplace = True)
result.sort_values(by=['RMSE'])


# In[151]:


#Get the overview distribution of the performance scores
fig, ((ax1, ax2, ax3), (ax4, ax5, ax6)) = plt.subplots(2,3,figsize=(16,12))
fig1 = result['MSE'].plot(kind='bar',title ="MSE", color = 'r',fontsize=12,ax=ax1)
fig2 = result['RMSE'].plot(kind='bar', title ="RMSE", color = 'b', fontsize=12,ax=ax2)
fig3 = result['R Squared'].plot(kind='bar', title ="R Squared", color = 'g', fontsize=12,ax=ax3)
fig4 = result['CV MSE mean'].plot(kind='bar', title ="CV MSE mean", color = 'c', fontsize=12,ax=ax4)
fig5 = result['CV MSE std'].plot(kind='bar', title ="CV MSE std",color = 'm', fontsize=12,ax=ax5)
fig6 = result['Model score'].plot(kind='bar', title ="Model score",color = 'y', fontsize=12,ax=ax6)
for ax in fig.axes:
    ax.tick_params(labelrotation=0)


# Ridge regression tended to give us a slightly better leaderboard score overall because it generates the lowest RMSE(MSE) score and highest R-Squared score which means have more predictive power to explain the variability within this model. Therefore, we decided to use Ridge regression as the final model for this project. In order to get better prediction results and there is room for further tuning and scaling to achieve more accurate prediction results. In the next step, we will also implement more regression and Neural Network models if time permits. 
